from abc import abstractmethod
from inspect import signature
from typing import Any, Callable, Generic, List, Sequence, Tuple, TypeVar

import pandas as pd
import torch as th
from torch.utils.data import DataLoader, Dataset

from ...utils.interfaces import UdaoInput, UdaoItemShape
from ..containers import TabularContainer

T = TypeVar("T")
ST = TypeVar("ST")


class BaseIterator(Dataset, Generic[T, ST]):
    """Base class for all dataset iterators.
    Inherits from torch.utils.data.Dataset.

    T is the type of the iterator output.
    ST is the type of the iterator output shape.

    See FeatureIterator for an example.
    """

    def __init__(self, keys: Sequence[str]) -> None:
        self.keys = keys
        self.tensors_dtype = th.float32
        self.augmentations: List[Callable[[T], T]] = []
        pass

    def __len__(self) -> int:
        return len(self.keys)

    @abstractmethod
    def _getitem(self, idx: int, /) -> T:
        pass

    def __getitem__(self, idx: int, /) -> T:
        """Returns the item at the given index."""
        item = self._getitem(idx)
        for augmentation in self.augmentations:
            item = augmentation(item)
        return item

    def set_augmentations(self, augmentations: List[Callable[[T], T]]) -> None:
        """Sets the augmentations to apply to the iterator output."""
        self.augmentations = augmentations

    @property
    @abstractmethod
    def shape(self) -> ST:
        """Returns the shape of the iterator output."""

    @staticmethod
    @abstractmethod
    def collate(items: List[T]) -> T:
        """Collates the items into a batch.
        Used in the dataloader."""
        pass

    def get_dataloader(
        self,
        batch_size: int,
        shuffle: bool = False,
        num_workers: int = 0,
        **kwargs: Any,
    ) -> DataLoader:
        """Returns a torch dataloader for the iterator,
        that can be used for training.
        This will use the collate static method
        to collate the items into a batch.
        """
        return DataLoader[T](
            self,
            batch_size=batch_size,
            shuffle=shuffle,
            collate_fn=self.collate,
            num_workers=num_workers,
            **kwargs,
        )

    @classmethod
    def get_parameter_names(cls) -> List[str]:
        """Returns the names of the container parameters
        of the iterator.
        Useful to create dynamic parameters for related parts
        of the pipeline (feature extractors, preprocessors)
        """
        init_signature = signature(cls.__init__)
        params_list = list(init_signature.parameters.keys())
        # remove self, args and kwargs
        params_list = [
            param
            for param in params_list
            if param not in ["self", "keys", "args", "kwargs"]
        ]
        return params_list

    def _get_sample(self) -> T:
        """Returns a random sample from the iterator."""
        return self[0]

    def set_tensors_dtype(self, dtype: th.dtype) -> None:
        """Sets the dtype of the iterator.
        Useful for mixed precision training.
        """
        self.tensors_dtype = dtype


# Type of the iterator output - in the Udao case,
# restricted to UdaoInput and its subclasses
UT = TypeVar("UT", bound=UdaoInput)
# Type of the iterator output shape - in the Udao case,
# restricted to UdaoInputShape and its subclasses
UST = TypeVar("UST", bound=UdaoItemShape)


class UdaoIterator(BaseIterator[Tuple[UT, th.Tensor], UST], Generic[UT, UST]):
    """Base iterator for the Udao use case, where the iterator
    returns a FeatureInput object. It is expected to accept:
    - a TabularContainer representing the tabular features
    which can be set as variables by the user in the optimization pipeline
    - a TabularContainer representing the objectives

    FST: Type of the iterator output shape - in the Udao case,
    restricted to FeatureInputShape and its subclasses.

    FT: Type of the iterator output - in the Udao case,
    restricted to th.Tensor and its subclasses
    This results in a type Tuple[UT, th.Tensor] for the iterator output.

    Parameters
    ----------
    keys : Sequence[str]
        Keys of the dataset, used for accessing all features
    tabular_features : TabularContainer
        Tabular features of the iterator
    objectives : TabularContainer
        Objectives of the iterator
    """

    def __init__(
        self,
        keys: Sequence[str],
        tabular_features: TabularContainer,
        objectives: TabularContainer,
    ) -> None:
        super().__init__(keys)
        self.tabular_features = tabular_features
        self.objectives = objectives

    def get_tabular_features_container(
        self, feature_input: th.Tensor
    ) -> TabularContainer:
        indices = [
            i
            for i, name in enumerate(self.shape.feature_names)
            if name in self.tabular_features.data.columns
        ]
        tabular_features = feature_input[:, indices]
        tabular_df = pd.DataFrame(
            tabular_features.cpu().numpy(), columns=self.tabular_features.data.columns
        )
        return TabularContainer(tabular_df)



from abc import ABC, abstractmethod
from typing import Optional, Sequence

import numpy as np


class BasePredicateEmbedder(ABC):
    @abstractmethod
    def fit_transform(
        self, training_texts: Sequence[str], epochs: Optional[int] = None
    ) -> np.ndarray:
        pass

    @abstractmethod
    def transform(self, texts: Sequence[str]) -> np.ndarray:
        pass



from typing import Any, Callable, Dict, Optional, Type, Union

import lightning.pytorch as pl
import pytorch_warmup as warmup
from torch.optim.lr_scheduler import CosineAnnealingLR, LRScheduler

from ..module import UdaoModule

SchedulerType = Union[LRScheduler, Callable[[pl.Trainer, UdaoModule], LRScheduler]]


class UdaoLRScheduler(pl.Callback):
    """Manual LR scheduler for pytorch-lightning.
    This is a callback that can be used to manually control the learning rate.
    It can be used to implement warmup or other learning rate schedules,
    in particular when needing access to data properties such
    as the number of training batches.

    Parameters
    ----------
    scheduler : SchedulerType
        Either a torch.optim.lr_scheduler.LRScheduler or
        a callable that instantiates a scheduler based on the trainer and module.
        This second option is useful when the scheduler needs access to data properties.
    warmup_cls : Type[warmup.BaseWarmup]
        A warmup class from pytorch-warmup. Will be instantiated with the optimizer.
        If None, no warmup will be used.
    scheduler_params : Optional[Dict[str, Any]], optional
        Parameters to pass at scheduler instantiation, by default None
    warmup_params : Optional[Dict[str, Any]], optional
        Parameters to pass at warmup instantiation, by default None
    """

    def __init__(
        self,
        scheduler: SchedulerType,
        warmup_cls: Type[warmup.BaseWarmup],
        scheduler_params: Optional[Dict[str, Any]] = None,
        warmup_params: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.scheduler_params = scheduler_params if scheduler_params else {}
        self.warmup_params = warmup_params if warmup_params else {}
        self.warmup_cls = warmup_cls
        self.scheduler = None
        self.scheduler_generator = None
        self.warmup = None
        if isinstance(scheduler, LRScheduler):
            self.scheduler = scheduler
        elif callable(scheduler):
            self.scheduler_generator = scheduler

    def on_train_start(self, trainer: pl.Trainer, module: UdaoModule) -> None:  # type: ignore
        if self.scheduler_generator:
            self.scheduler = self.scheduler_generator(
                trainer, module, **self.scheduler_params
            )
        if self.warmup_cls:
            self.warmup = self.warmup_cls(trainer.optimizers[0], **self.warmup_params)
        # Assuming the optimizer has already been configured in pl_module

    def on_train_batch_end(self, *args: Any) -> None:
        if self.scheduler is None:
            raise ValueError("Scheduler not initialized")
        if self.warmup is not None:
            with self.warmup.dampening():
                self.scheduler.step()
        else:
            self.scheduler.step()

    def on_save_checkpoint(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        checkpoint: Dict[str, Any],
    ) -> None:
        checkpoint["scheduler_state_dict"] = (
            self.scheduler.state_dict() if self.scheduler else None
        )

    def on_load_checkpoint(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        checkpoint: Dict[str, Any],
    ) -> None:
        if self.scheduler:
            self.scheduler.load_state_dict(checkpoint["scheduler_state_dict"])


def setup_cosine_annealing_lr(
    trainer: pl.Trainer, pl_module: UdaoModule
) -> CosineAnnealingLR:
    """Instantiation of a CosineAnnealingLR scheduler
    from data size and number of epochs."""
    num_training_batches = len(trainer.train_dataloader)  # type: ignore
    max_steps = 0
    if trainer.max_steps:
        max_steps = trainer.max_steps
    elif trainer.max_epochs:
        max_steps = num_training_batches * trainer.max_epochs
    else:
        raise ValueError("Either max_steps or max_epochs must be set")
    optimizer = trainer.optimizers[0]
    scheduler = CosineAnnealingLR(
        optimizer,
        T_max=int(max_steps),
        eta_min=pl_module.learning_params.min_lr,
    )
    return scheduler



from typing import Any, cast

import dgl
import pytest
import torch as th

from ...utils.query_plan import (
    QueryPlanOperationFeatures,
    QueryPlanStructure,
    add_positional_encoding,
    compute_meta_features,
    extract_query_plan_features,
    get_laplacian_positional_encoding,
    random_flip_positional_encoding,
)
from ..conftest import QueryPlanElements


# Define a custom fixture to handle both sample plans
@pytest.fixture(params=["sample_plan_1", "sample_plan_2"])
def query_plan_sample(request: Any) -> QueryPlanElements:
    return request.getfixturevalue(request.param)


def test_logical_struct(
    query_plan_sample: QueryPlanElements,
) -> None:
    expected_structure = query_plan_sample.structure
    expected_op_features = query_plan_sample.features
    expected_meta_features = query_plan_sample.meta_features
    query = query_plan_sample.query_plan
    structure, features, meta_features = extract_query_plan_features(query)
    assert features.size == expected_op_features.size
    assert features.rows_count == expected_op_features.rows_count
    assert meta_features["meta_size"] == expected_meta_features["meta_size"]
    assert meta_features["meta_rows_count"] == expected_meta_features["meta_rows_count"]
    assert structure.incoming_ids == expected_structure.incoming_ids
    assert structure.outgoing_ids == expected_structure.outgoing_ids
    assert structure.node_id2name == expected_structure.node_id2name


def test_compare_same_struct_is_True(sample_plan_1: QueryPlanElements) -> None:
    structure, _, _ = extract_query_plan_features(sample_plan_1.query_plan)
    assert structure.graph_match(structure) is True


def test_compare_different_structures_is_False(
    sample_plan_1: QueryPlanElements, sample_plan_2: QueryPlanElements
) -> None:
    structure, _, _ = extract_query_plan_features(sample_plan_1.query_plan)
    structure_2, _, _ = extract_query_plan_features(sample_plan_2.query_plan)
    assert structure.graph_match(structure_2) is False


def test_compute_meta_features_one_node() -> None:
    structure = QueryPlanStructure(["a", "b", "c"], [0, 1], [1, 2])
    features = QueryPlanOperationFeatures([1.0, 2.0, 3.0], [10, 20, 30])
    meta_features = compute_meta_features(structure, features)
    assert meta_features["meta_size"] == 10
    assert meta_features["meta_rows_count"] == 1.0


def test_compute_meta_features_two_nodes() -> None:
    structure = QueryPlanStructure(
        ["a", "b", "c", "d", "e", "f"], [0, 1, 3, 4], [1, 2, 4, 5]
    )
    features = QueryPlanOperationFeatures(
        [1.0, 2.0, 3.0, 4.0, 5.0, 6.0], [10, 20, 30, 40, 50, 60]
    )
    meta_features = compute_meta_features(structure, features)
    assert meta_features["meta_size"] == 50
    assert meta_features["meta_rows_count"] == 5


def test_get_laplacian_positional_encoding() -> None:
    graph = dgl.graph(([0, 1], [1, 2]))
    encoding = get_laplacian_positional_encoding(graph, 1)
    assert encoding.shape == (3, 1)
    assert th.allclose(encoding, th.tensor([[0.0], [0.0], [1.0]]))


def test_add_positional_encoding() -> None:
    graph = dgl.graph(([0, 1], [1, 2]))
    graph = add_positional_encoding(graph, 1)
    result = cast(th.Tensor, graph.ndata["pos_enc"])
    assert result.shape == (3, 1)
    assert th.allclose(result, th.tensor([[-7.0711e-01], [0.0], [7.0711e-01]]))


def test_random_flip_positional_encoding() -> None:
    graph = dgl.graph(([0, 1], [1, 2]))
    th.manual_seed(0)
    positional_encoding = th.tensor(
        [[1, 1, 1], [-1, -1, 1], [-1, 1, 1]], dtype=th.float32
    )
    graph.ndata["pos_enc"] = positional_encoding
    return_graph = random_flip_positional_encoding(graph)
    return_encoding = cast(th.Tensor, return_graph.ndata["pos_enc"])
    # same flipping for each node
    assert th.equal(
        return_encoding,
        th.tensor([[-1.0, 1.0, -1.0], [1, -1, -1], [1, 1, -1]]),
    )



import itertools
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import torch as th
from torch.multiprocessing import Pool

from ....utils.logging import logger
from ...concepts import MOProblem, Objective, SOProblem
from ...soo.so_solver import SOSolver
from ...utils import moo_utils as moo_ut
from ...utils.exceptions import NoSolutionError
from ...utils.moo_utils import Point, Rectangle
from .base_progressive_frontier import BaseProgressiveFrontier


class ParallelProgressiveFrontier(BaseProgressiveFrontier):
    @dataclass
    class Params(BaseProgressiveFrontier.Params):
        processes: int = 1
        """Processes to use for parallel processing"""
        n_grids: int = 2
        """Number of splits per objective"""
        max_iters: int = 10
        """Number of iterations to explore the space"""

    def __init__(
        self,
        solver: SOSolver,
        params: Params,
    ) -> None:
        super().__init__(
            solver,
            params,
        )
        self.processes = params.processes
        self.n_grids = params.n_grids
        self.max_iters = params.max_iters

    def solve(
        self,
        problem: MOProblem,
        seed: Optional[int] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        solve MOO by PF-AP (Progressive Frontier - Approximation Parallel)

        Parameters
        ----------
        problem : MOProblem
            MOO problem to be solved
        Returns
        -------
        po_objs: ndarray
            Pareto optimal objective values, of shape
            (n_solutions, n_objs)
        po_vars: ndarray
            corresponding variables of Pareto solutions, of shape
            (n_solutions, n_vars)
        """
        # create initial rectangle
        # get initial plans/form a intial hyperrectangle

        plans: List[Point] = []
        n_objs = len(problem.objectives)

        all_objs_list: List[np.ndarray] = []
        all_vars_list: List[Dict] = []
        for i in range(n_objs):
            anchor_point = self.get_anchor_point(problem=problem, obj_ind=i, seed=seed)
            if anchor_point.vars is None:
                raise Exception("This should not happen.")
            plans.append(anchor_point)
            all_objs_list.append(anchor_point.objs)
            all_vars_list.append(anchor_point.vars)
        logger.debug(f"the initial plans are: {plans}")
        if n_objs < 2 or n_objs > 3:
            raise Exception(f"{n_objs} objectives are not supported for now!")

        for i in range(self.max_iters):
            # choose the cell with max volume to explore
            max_volume = -1
            input_ind = -1
            for i in range(len(all_objs_list) - 1):
                current_volume = abs(
                    np.prod(np.array(all_objs_list)[i] - np.array(all_objs_list)[i + 1])
                )
                logger.debug(f"volume {current_volume}")
                if current_volume > max_volume:
                    max_volume = current_volume
                    input_ind = i

            plan = [
                Point(objs=np.array(all_objs_list)[input_ind]),
                Point(objs=np.array(all_objs_list)[input_ind + 1]),
            ]
            utopia, nadir = self.get_utopia_and_nadir(plan)
            if utopia is None or nadir is None:
                raise NoSolutionError("Cannot find utopia/nadir points")
            # create uniform n_grids ^ (n_objs) grid cells based on the rectangle
            grid_cells_list = self._create_grid_cells(
                utopia, nadir, self.n_grids, n_objs
            )

            obj_bound_cells = []
            for cell in grid_cells_list:
                obj_bound_dict = self._form_obj_bounds_dict(
                    problem, cell.utopia, cell.nadir
                )
                obj_bound_cells.append(obj_bound_dict)

            logger.debug(f"the cells are: {obj_bound_cells}")
            ret_list = self.parallel_soo(
                problem=problem,
                objective=problem.objectives[self.opt_obj_ind],
                cell_list=obj_bound_cells,
                seed=seed,
            )

            po_objs_list: List[np.ndarray] = []
            po_vars_list: List[Dict] = []
            for soo_obj, soo_vars in ret_list:
                if soo_obj is None:
                    logger.debug("This is an empty area!")
                    continue
                else:
                    po_objs_list.append(self._compute_objectives(problem, soo_vars))
                    if soo_vars is None:
                        raise Exception("Unexpected vars None for objective value.")
                    po_vars_list.append(soo_vars)

            logger.debug(f"the po_objs_list is: {po_objs_list}")
            logger.debug(f"the po_vars_list is: {po_vars_list}")
            all_objs_list.extend(po_objs_list)
            all_vars_list.extend(po_vars_list)
            logger.debug(f"the all_objs_list is: {all_objs_list}")
            all_objs, all_vars = moo_ut.summarize_ret(all_objs_list, all_vars_list)
            all_objs_list = all_objs.tolist() if all_objs is not None else []
            all_vars_list = all_vars.tolist() if all_vars is not None else []

        return np.array(all_objs_list), np.array(all_vars_list)

    def _solve_wrapper(
        self, problem: SOProblem, seed: Optional[int] = None
    ) -> Optional[Tuple[float, Dict[str, Any]]]:
        """Handle exceptions in solver call for parallel processing."""
        try:
            return self.solver.solve(problem, seed=seed)
        except NoSolutionError:
            logger.debug(f"This is an empty area! {problem}")
            return None

    def parallel_soo(
        self,
        problem: MOProblem,
        objective: Objective,
        cell_list: List[Dict[str, Any]],
        seed: Optional[int] = None,
    ) -> List[Tuple[float, Dict[str, Any]]]:
        """Parallel calls to SOO Solver for each cell in cell_list, returns a
        candidate tuple (objective_value, variables) for each cell.

        Parameters
        ----------
        objective : Objective
            Objective to be optimized
        cell_list : List[Dict[str, Any]]
            List of cells to be optimized
            (a cell is a dict of bounds for each objective)
        input_parameters : Optional[Dict[str, Any]], optional
            Fixed parameters to be passed , by default None

        Returns
        -------
        List[Tuple[float, Dict[str, Any]]]
            List of candidate tuples (objective_value, variables)
        """
        # generate the list of input parameters for constraint_so_opt
        args_list: List[Tuple[SOProblem, Optional[int]]] = []
        for obj_bounds_dict in cell_list:
            so_problem = self._so_problem_from_bounds_dict(
                problem, obj_bounds_dict, objective
            )
            args_list.append((so_problem, seed))

        if th.cuda.is_available():
            th.multiprocessing.set_start_method("spawn", force=True)
        if self.processes == 1:
            ret_list = [self._solve_wrapper(*args) for args in args_list]
        else:
            # call self.constraint_so_opt parallely
            with Pool(processes=self.processes) as pool:
                ret_list = pool.starmap(self._solve_wrapper, args_list)
        return [res for res in ret_list if res is not None]

    @staticmethod
    def _create_grid_cells(
        utopia: Point, nadir: Point, n_grids: int, n_objs: int
    ) -> List[Rectangle]:
        """
        Create cells used in Progressive Frontier(PF)-Approximation
        Parallel (AP) algorithm

        Parameters
        ----------
        utopia: Point
            the utopia point
        nadir: Point
            the nadir point
        n_grids: int
            the number of grids per objective
        n_objs: int
            the number of objectives

        Returns
        -------
            List[Rectangle]
            The rectangles in which to perform optimization.
        """
        grids_per_var = np.linspace(
            utopia.objs, nadir.objs, num=n_grids + 1, endpoint=True
        )
        objs_list = [grids_per_var[:, i] for i in range(n_objs)]

        ## generate cartesian product of indices for grids
        grids_inds_per_var = np.linspace(0, n_grids - 1, num=n_grids, endpoint=True)
        x = np.tile(grids_inds_per_var, (n_objs, 1))
        grids_inds = np.array([list(i) for i in itertools.product(*x)]).astype(int)

        grid_cell_list = []
        for grid_ind in grids_inds:
            sub_u_objs = np.array([objs_list[i][id] for i, id in enumerate(grid_ind)])
            sub_u_point = Point(sub_u_objs)
            sub_nadir_objs = np.array(
                [objs_list[i][id + 1] for i, id in enumerate(grid_ind)]
            )
            sub_nadir_point = Point(sub_nadir_objs)
            assert all((sub_nadir_objs - sub_u_objs) >= 0)
            cell = Rectangle(sub_u_point, sub_nadir_point)
            grid_cell_list.append(cell)
        if len(grid_cell_list) != (n_grids**n_objs):
            raise Exception(
                f"Unexpected: the number of grid cells is"
                f"not equal to {n_grids**n_objs}"
            )

        return grid_cell_list



import pandas as pd
import pytest

from ...predicate_embedders.utils import extract_operations, prepare_operation


@pytest.mark.parametrize(
    "op,expected",
    [
        (
            "GlobalLimit 10, Statistics(sizeInBytes=400.0 B, rowCount=10)",
            "globallimit 10",
        ),
        (
            "+- LocalLimit 10, Statistics(sizeInBytes=6.9 GiB, rowCount=1.86E+8)",
            "locallimit 10",
        ),
        (
            "   +- Sort [revenue#47608 DESC NULLS LAST, "
            "o_orderdate#22 ASC NULLS FIRST], true, Statistics(sizeInBytes=6.9 GiB,"
            " rowCount=1.86E+8)",
            "sort revenue desc nulls last o_orderdate asc nulls first true",
        ),
        (
            "      +- Aggregate [l_orderkey#23L, o_orderdate#22,"
            " o_shippriority#20], [l_orderkey#23L, sum(CheckOverflow"
            "((promote_precision(cast(l_extendedprice#28 as "
            "decimal(13,2))) * promote_precision(CheckOverflow((1.00"
            " - promote_precision(cast(l_discount#29 as decimal(13,2))))"
            ", DecimalType(13,2), true))), DecimalType(26,4), true)) AS"
            " revenue#47608, o_orderdate#22, o_shippriority#20],"
            " Statistics(sizeInBytes=6.9 GiB, rowCount=1.86E+8)",
            "aggregate l_orderkey o_orderdate o_shippriority l_orderkey"
            " sum checkoverflow promote_precision cast l_extendedprice as"
            " decimal 13 2 rmul promote_precision checkoverflow 1.00 rminus"
            " promote_precision cast l_discount as decimal 13 2 decimaltype 13 2"
            " true decimaltype 26 4 true as revenue o_orderdate o_shippriority",
        ),
        (
            "               +- Relation tpch_100.part[p_partkey#59113L,p_name#59114"
            ",p_mfgr#59115,p_type#59116,p_size#59117,p_container#59118,p_retailprice"
            "#59119,p_comment#59120,p_brand#59121] parquet, Statistics(sizeInBytes=3.5"
            " GiB, rowCount=1.92E+7)",
            "relation tpch_100.part p_partkey p_name p_mfgr p_type p_size p_container"
            " p_retailprice p_comment p_brand parquet",
        ),
    ],
)
def test_prepare_operation(op: str, expected: str) -> None:
    prepared = prepare_operation(op)
    assert prepared == expected


@pytest.fixture
def df_fixture() -> pd.DataFrame:
    return pd.DataFrame.from_dict(
        {
            "id": [1, 2],
            "plan": [
                "a b c\na b d",
                "a b d\na b x",
            ],
        }
    )


def test_extract_operations(df_fixture: pd.DataFrame) -> None:
    plan_to_op, operations = extract_operations(df_fixture)
    assert plan_to_op == {1: [0, 1], 2: [1, 2]}
    assert operations == ["a b c", "a b d", "a b x"]


def test_extract_operations_processing_is_applied(df_fixture: pd.DataFrame) -> None:
    plan_to_op, operations = extract_operations(
        df_fixture, operation_processing=lambda s: s.replace("x", "c")
    )
    assert plan_to_op == {1: [0, 1], 2: [1, 0]}
    assert operations == ["a b c", "a b d"]



from abc import ABC, abstractmethod
from dataclasses import dataclass

import torch as th
from torch import nn


class BaseRegressor(nn.Module, ABC):
    @dataclass
    class Params:
        input_embedding_dim: int
        """Size of the embedding part of the input."""
        input_features_dim: int  # depends on the data
        """Size of the tabular features."""
        output_dim: int
        """Size of the output tensor."""

    def __init__(self, net_params: Params) -> None:
        super().__init__()
        self.input_dim = net_params.input_embedding_dim + net_params.input_features_dim
        self.output_dim = net_params.output_dim

    @abstractmethod
    def forward(self, embedding: th.Tensor, inst_feat: th.Tensor) -> th.Tensor:
        ...



from typing import cast

import numpy as np
import pandas as pd
import pytest
import torch as th

from ...extractors import QueryStructureExtractor
from ...utils.query_plan import QueryPlanStructure
from ..conftest import QueryPlanElements


@pytest.fixture
def df_fixture(
    sample_plan_1: QueryPlanElements, sample_plan_2: QueryPlanElements
) -> pd.DataFrame:
    input_df = pd.DataFrame.from_dict(
        {
            "id": [1, 2, 3],
            "tid": [1, 2, 1],
            "plan": [
                sample_plan_1.query_plan,
                sample_plan_2.query_plan,
                sample_plan_1.query_plan,
            ],
        }
    )
    return input_df


class TestStructureExtractor:
    def test_structure_extractor_has_feature_types(self) -> None:
        extractor = QueryStructureExtractor()
        assert extractor.feature_types == {"rows_count": float, "size": float}

    def test_structures_match_templates(self, df_fixture: pd.DataFrame) -> None:
        extractor = QueryStructureExtractor()
        for row in df_fixture.itertuples():
            s_dict = extractor._extract_structure_and_features(
                row.id, row.plan, "train"
            )
            assert set(s_dict.keys()) == {
                "operation_id",
                "operation_gid",
                *extractor.feature_types.keys(),
                *[f"meta_{feature}" for feature in extractor.feature_types.keys()],
            }
            for key, value in s_dict.items():
                if "meta" in key:
                    assert isinstance(value, float)
                else:
                    assert len(value) == len(row.plan.splitlines())
        for plan in extractor.template_plans.values():
            assert type(plan) == QueryPlanStructure
        assert len(extractor.template_plans) == 2
        assert extractor.id_template_dict == {1: 1, 2: 2, 3: 1}

    def test_extract_structure_from_df_returns_correct_shape(
        self, df_fixture: pd.DataFrame
    ) -> None:
        """Graph features and meta features have the correct shape"""
        extractor = QueryStructureExtractor()
        structure_container = extractor.extract_features(df_fixture, "train")

        multi_index = pd.MultiIndex.from_tuples(
            [
                [row.id, i]
                for row in df_fixture.itertuples()
                for i, _ in enumerate(row.plan.splitlines())
            ],
            names=["plan_id", "operation_id"],
        )
        assert (
            graph_meta_features := structure_container.graph_meta_features
        ) is not None
        assert graph_meta_features.shape == (len(df_fixture), 2)

        np.testing.assert_array_equal(
            graph_meta_features.columns, ["rows_count", "size"]
        )
        assert (multi_index == structure_container.graph_features.index).all()

    def test_extract_structure_from_unseen_structure(
        self, df_fixture: pd.DataFrame
    ) -> None:
        """Extracting features from a query plan that is not in the training set"""
        extractor = QueryStructureExtractor()
        structure_container = extractor.extract_features(df_fixture, "val")
        assert len(structure_container.template_plans) == 2
        assert structure_container.graph_meta_features is not None
        assert structure_container.graph_meta_features.shape == (len(df_fixture), 2)

    def test_extract_structure_from_df_returns_correct_values(
        self, df_fixture: pd.DataFrame
    ) -> None:
        """Values in the graph_features and graph_meta_features dataframes
        match values in the dictionary"""
        extractor = QueryStructureExtractor()
        structure_container = extractor.extract_features(df_fixture, "train")
        for row in df_fixture.itertuples():
            features_dict = extractor._extract_structure_and_features(
                row.id, row.plan, "val"
            )
            for feature in ["rows_count", "size"]:
                assert (
                    graph_meta_features := structure_container.graph_meta_features
                ) is not None
                np.testing.assert_allclose(
                    structure_container.graph_features.loc[row.id][feature].values,
                    features_dict[feature],
                    rtol=1e-6,
                )
                assert structure_container.graph_meta_features is not None
                np.testing.assert_allclose(
                    graph_meta_features.loc[row.id][feature],
                    features_dict[f"meta_{feature}"],
                    rtol=1e-6,
                )

    def test_extract_operation_types(self, df_fixture: pd.DataFrame) -> None:
        extractor = QueryStructureExtractor()
        row = df_fixture.iloc[1]
        # avoid error message for template id.
        extractor.extract_features(df_fixture, "train")
        extractor.operation_types = {}
        features_dict = extractor._extract_structure_and_features(
            row.id, row.plan, "train"
        )
        assert features_dict["operation_gid"] == [0, 1, 2, 3, 4]

        row = df_fixture.iloc[0]
        features_dict = extractor._extract_structure_and_features(
            row.id, row.plan, "val"
        )
        assert features_dict["operation_gid"] == [
            -1,
            -1,
            0,
            1,
            2,
            -1,
            2,
            -1,
            2,
            -1,
            2,
            3,
            4,
            2,
            3,
            4,
            2,
            3,
            4,
            2,
            3,
            4,
        ]
        features_dict = extractor._extract_structure_and_features(
            row.id, row.plan, "train"
        )

        assert features_dict["operation_gid"] == [
            5,
            6,
            0,
            1,
            2,
            7,
            2,
            7,
            2,
            7,
            2,
            3,
            4,
            2,
            3,
            4,
            2,
            3,
            4,
            2,
            3,
            4,
        ]

    @pytest.mark.parametrize("positional_encoding_size", [2, 3])
    def test_extract_structure_with_pe(
        self, df_fixture: pd.DataFrame, positional_encoding_size: int
    ) -> None:
        extractor = QueryStructureExtractor(
            positional_encoding_size=positional_encoding_size
        )
        row = df_fixture.iloc[1]
        extractor._extract_structure_and_features(row.id, row.plan, "train")
        result = cast(th.Tensor, extractor.template_plans[1].graph.ndata["pos_enc"])
        th.allclose(
            result,
            th.tensor(
                [
                    [-0.4508, 1.0000, 0.4508],
                    [-0.6977, 0.0000, -0.6977],
                    [-0.5091, 0.0000, 0.5091],
                    [0.0563, 0.0000, 0.0563],
                    [0.2181, 0.0000, -0.2181],
                ]
            )[:, :positional_encoding_size],
        )

    def test_extract_structure_with_pe_at_0(self, df_fixture: pd.DataFrame) -> None:
        extractor = QueryStructureExtractor(positional_encoding_size=0)
        row = df_fixture.iloc[1]
        extractor._extract_structure_and_features(row.id, row.plan, "train")
        assert "pos_enc" not in extractor.template_plans[1].graph.ndata



import pytest
import torch as th

from ...embedders.base_graph_embedder import BaseGraphEmbedder
from .conftest import generate_dgl_graph


def test_base_embedder_initialization() -> None:
    params = BaseGraphEmbedder.Params(
        input_size=5,
        output_size=10,
        op_groups=["type", "cbo"],
        type_embedding_dim=5,
        embedding_normalizer="BN",
        n_op_types=3,
    )

    embedder = BaseGraphEmbedder(params)

    assert embedder.input_size == 10
    assert embedder.embedding_size == 10
    assert embedder.op_type
    assert embedder.op_cbo
    assert not embedder.op_enc


def test_base_embedder_invalid_normalizer() -> None:
    params = BaseGraphEmbedder.Params(
        input_size=5,
        output_size=10,
        op_groups=["type", "cbo"],
        type_embedding_dim=5,
        embedding_normalizer="UNKNOWN",  # type: ignore
        n_op_types=3,
    )

    with pytest.raises(ValueError):
        BaseGraphEmbedder(params)


def test_base_embedder_concatenate_op_features() -> None:
    params = BaseGraphEmbedder.Params(
        input_size=5,
        output_size=10,
        op_groups=["type", "cbo"],
        type_embedding_dim=5,
        embedding_normalizer=None,
        n_op_types=3,
    )

    embedder = BaseGraphEmbedder(params)
    g = generate_dgl_graph(
        3,
        2,
        {"op_gid": {"size": 1, "type": "int"}, "cbo": {"size": 2, "type": "float"}},
    )
    result = embedder.concatenate_op_features(g)

    assert isinstance(result, th.Tensor)
    assert result.shape == (3, 5 + 2)  # 5 for embedding and 2 for "cbo"


def test_base_embedder_normalize_embedding() -> None:
    params = BaseGraphEmbedder.Params(
        input_size=5,
        output_size=3,
        op_groups=["type"],
        type_embedding_dim=5,
        embedding_normalizer="BN",
        n_op_types=3,
    )

    embedder = BaseGraphEmbedder(params)
    tensor = th.randn(5, 3)

    normalized_tensor = embedder.normalize_embedding(tensor)

    assert isinstance(normalized_tensor, th.Tensor)
    assert normalized_tensor.shape == tensor.shape



import heapq
import itertools
from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np

from ....utils.logging import logger
from ...concepts.problem import MOProblem
from ...soo.so_solver import SOSolver
from ...utils import moo_utils as moo_ut
from ...utils.exceptions import NoSolutionError
from ...utils.moo_utils import Point, Rectangle
from .base_progressive_frontier import BaseProgressiveFrontier


class SequentialProgressiveFrontier(BaseProgressiveFrontier):
    """
    Sequential Progressive Frontier -
    a progressive frontier algorithm that explores the uncertainty space
    sequentially, by dividing the space into subrectangles and
    exploring the subrectangles one by one.
    """

    @dataclass
    class Params(BaseProgressiveFrontier.Params):
        n_probes: int = 10
        """number of probes"""

    def __init__(
        self,
        solver: SOSolver,
        params: Params,
    ) -> None:
        super().__init__(solver, params)
        self.n_probes = params.n_probes

    def solve(
        self,
        problem: MOProblem,
        seed: Optional[int] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Solve MOO by Progressive Frontier

        Parameters
        ----------
        problem : MOProblem
            MOO problem to be solved

        Returns
        -------
        Tuple[np.ndarray | None, np.ndarray | None]
            optimal objectives and variables
            None, None if no solution is found
        """
        rectangle_queue: List[Rectangle] = []
        n_objs = len(problem.objectives)
        plans = [
            self.get_anchor_point(problem=problem, obj_ind=i, seed=seed)
            for i in range(n_objs)
        ]
        utopia, nadir = self.get_utopia_and_nadir(plans)
        rectangle = Rectangle(utopia, nadir)
        heapq.heappush(rectangle_queue, rectangle)
        for _ in range(self.n_probes - n_objs):
            if not rectangle_queue:
                logger.info("No more uncertainty space to explore further!")
                break
            rectangle = heapq.heappop(rectangle_queue)
            middle_point, subrectangles = self._find_local_optimum(
                problem, rectangle, seed=seed
            )
            if middle_point is not None:
                plans.append(middle_point)
            for sub_rect in subrectangles:
                if sub_rect.volume != 0:
                    heapq.heappush(rectangle_queue, sub_rect)

        ## filter dominated points
        po_objs_list = [point.objs.tolist() for point in plans]
        po_vars_list = [point.vars for point in plans]
        po_objs, po_vars = moo_ut.summarize_ret(po_objs_list, po_vars_list)

        return po_objs, po_vars

    def _find_local_optimum(
        self,
        problem: MOProblem,
        rectangle: Rectangle,
        seed: Optional[int] = None,
    ) -> Tuple[Optional[Point], List[Rectangle]]:
        """
        Find the local optimum in the given rectangle and
        the subrectangles in which to continue the search.
        If no optimum is found for the rectangle,
        return None and the upper subrectangles.
        If an optimum is found, return the optimum and
        the subrectangles in which to continue the search.

        Parameters
        ----------
        rectangle : Rectangle
            Rectangle in which to find the local optimum
        wl_id : str | None
            workload id

        Returns
        -------
        Tuple[Point | None, List[Rectangle]]
            The local optimum and the subrectangles in which to continue the search
        """
        current_utopia, current_nadir = Point(rectangle.lower_bounds), Point(
            rectangle.upper_bounds
        )
        logger.debug(f"current utopia is: {current_utopia}")
        logger.debug(f"current nadir is: {current_nadir}")
        middle_objs = np.array(
            [
                current_nadir.objs[i]
                if i == self.opt_obj_ind
                else (current_utopia.objs[i] + current_nadir.objs[i]) / 2
                for i in range(len(problem.objectives))
            ]
        )
        middle_point = Point(middle_objs)
        obj_bounds_dict = self._form_obj_bounds_dict(
            problem, current_utopia, middle_point
        )
        logger.debug(f"obj_bounds are: {obj_bounds_dict}")
        so_problem = self._so_problem_from_bounds_dict(
            problem, obj_bounds_dict, problem.objectives[self.opt_obj_ind]
        )
        try:
            _, soo_vars = self.solver.solve(so_problem, seed=seed)
        except NoSolutionError:
            logger.debug(
                "This is an empty area \n "
                "don't have pareto points, only "
                "divide current uncertainty space"
            )
            middle_point = Point((current_utopia.objs + current_nadir.objs) / 2)
            rectangles = self.generate_sub_rectangles(
                current_utopia, current_nadir, middle_point, successful=False
            )
            return None, rectangles
        else:
            middle_objs = self._compute_objectives(problem, soo_vars)
            middle_point = Point(middle_objs, soo_vars)
            return middle_point, self.generate_sub_rectangles(
                current_utopia, current_nadir, middle_point
            )

    def generate_sub_rectangles(
        self, utopia: Point, nadir: Point, middle: Point, successful: bool = True
    ) -> List[Rectangle]:
        """

        Generate uncertainty space to be explored:
        - if starting from successful optimum as middle, excludes the dominated
        space (middle as utopia and nadir as nadir)
        - if starting from unsuccessful optimum as middle, excludes the space where
        all constraining objectives are lower than the middle point.

        Parameters
        ----------
        utopia: Point
            the utopia point
        nadir: Point
            the nadir point
        middle: Point
            the middle point generated by
            the constrained single objective optimization
        successful: bool
            whether the middle point is from a successful optimization

        Returns
        -------
        List[Rectangle]
            sub rectangles to be explored
        """

        rectangles = []
        corner_points = self._get_corner_points(utopia, nadir)
        for point in corner_points:
            # space explored (lower half of constraining objectives)
            is_explored_unconclusive = not successful and np.all(
                middle.objs[1:] - point.objs[1:] > 0
            )
            # nadir point
            is_dominated = successful and np.all(middle.objs - point.objs < 0)
            if is_dominated or is_explored_unconclusive:
                continue
            sub_rect_u, sub_rect_n = self.get_utopia_and_nadir([point, middle])
            rectangles.append(Rectangle(sub_rect_u, sub_rect_n))

        return rectangles

    def _get_corner_points(self, utopia: Point, nadir: Point) -> List[Point]:
        """
        get the corner points that can form a hyper_rectangle
        from utopia and nadir points.

        Parameters
        ----------
        utopia: Points (defined by class), the utopia point
        nadir: Points (defined by class), the nadir point

        Returns
        -------
        List[Point]
            2^n_objs corner points
        """
        n_objs = utopia.n_objs
        u_obj_values, n_obj_values = utopia.objs.reshape(
            [n_objs, 1]
        ), nadir.objs.reshape([n_objs, 1])
        grids_list = np.hstack([u_obj_values, n_obj_values])

        ## generate cartesian product of grids_list
        objs_corner_points = np.array([list(i) for i in itertools.product(*grids_list)])
        corner_points = [Point(objs=obj_values) for obj_values in objs_corner_points]

        return corner_points
